<?php	
	
	//Suppress warnings. Error messaging is handled explicitly.
	error_reporting(E_ERROR);
	
	//Usage: php example.xml 3000 3000 70 65.15506 25.177116 64.922305 25.732355 out.png
	$xmlFile = $argv[1];
	$width = intval($argv[2]);
	$height = intval($argv[3]);
	$alpha = intval($argv[4]);
	$topLat = floatval($argv[5]);	
	$leftLng = floatval($argv[6]);
	$bottomLat = floatval($argv[7]);
	$rightLng = floatval($argv[8]);	
	$imgOut = $argv[9];
	
	//Check parameters
	if(!isset($xmlFile)) die('XML filename missing. '.getUsageString().PHP_EOL);
	if(!isset($width)) die('Width missing. '.getUsageString().PHP_EOL);
	if(!isset($height)) die('Height missing. '.getUsageString().PHP_EOL);
	if(!isset($alpha)) die('Translucency missing. '.getUsageString().PHP_EOL);
	if(!isset($topLat)) die('Top latitude missing. '.getUsageString().PHP_EOL);
	if(!isset($leftLng)) die('Left longitude missing. '.getUsageString().PHP_EOL);
	if(!isset($bottomLat)) die('Bottom latitude missing. '.getUsageString().PHP_EOL);
	if(!isset($rightLng)) die('Right longitude missing. '.getUsageString().PHP_EOL);
	if(!isset($imgOut)) die('Image output filename missing. '.getUsageString().PHP_EOL);		
	if($width < 1 || $height < 1) die('Width and height must be positive. '.PHP_EOL);		
	if($alpha < 0 || $alpha > 127) die('Translucency value must be between 0 (opaque) and 127 (transparent). '.PHP_EOL);		
	
	//Load and validate the input XML
	$xmlDom = new DOMDocument();
	if(!$xmlDom->load($xmlFile)) {
		die("Could not read input XML from \"$xmlFile\".".PHP_EOL);
	}
	if (!($xmlDom->schemaValidateSource(getSchemaString()))) {
		die("Invalid input XML. See the readme file for correct format.".PHP_EOL);
	}
	
	//Convert the DOM object into a simple XML object for convenience
	$dataXML = simplexml_import_dom($xmlDom);
	
	//Create the heat map array (initialize y values only when necessary) 
	$heatMap = array();
	for($x = 0; $x < $width; $x++) {		
		$heatMap[$x] = array();
	}	
	
	//Map location coordinates to pixels and update user amounts
	$lngWidth = $rightLng - $leftLng;
	$latHeight = $topLat - $bottomLat;
	$mostUsers = 0;
	$totalUsers = 0;

	echo 'Generating the heat map image...'.PHP_EOL;	
	foreach ($dataXML->locations->location as $location) {
		$x = floor($width * ((floatval($location->longitude) - $leftLng) / $lngWidth));
		$y = floor($height * (($topLat - floatval($location->latitude)) / $latHeight));
		
		//Ignore stations that are outside the borders  
		if($x < 0 || $x >= $width || $y < 0 || $y >= $height) {
			continue;
		}

		$users = 0;
		if(intval($location->users) > 0) {			
			if(!isset($heatMap[$x][$y])) $heatMap[$x][$y] = 0;
			$users = intval($heatMap[$x][$y])+intval($location->users);
			$heatMap[$x][$y] = $users;
		}		

		if($users > $mostUsers) {
			$mostUsers = $users;
		}				
	}
	
	$maxDropPerPixel = 9999;
	if($mostUsers > 0) {
		$maxDropPerPixel = $mostUsers / 14.0;		
	}
	
	$maxDropPerPixelDiag = 9999;
	if($mostUsers > 0) {
		$maxDropPerPixelDiag = $mostUsers / 9.9;		
	}
	
		
	//Create slopes 
	for($y = 0; $y < $height; $y++) {		 
		for($x = 0; $x < $width; $x++) {
			if(isset($heatMap[$x][$y]) && intval($heatMap[$x][$y]) > 0) smoothenDrops($x, $y);
		}
	}	
	
	//Create the image	
	$img = imagecreatetruecolor($width, $height);
	imagealphablending($img, false);
	imagesavealpha($img, true);			
	
	//Color the pixels	
	for($y = 0; $y < $height; $y++) {		 
		for($x = 0; $x < $width; $x++) {
			//Set heat value between 0 and 1000
			if(isset($heatMap[$x][$y]) && $mostUsers > 0) {
				$heatValue = intval(1000 * (intval($heatMap[$x][$y])/$mostUsers));	
			}
			else {
				$heatValue = 0;
			}			
			
			$red = 0;
			$green = 0;
			$blue = 0;		
							
			if($heatValue < 25) {
				$pct = $heatValue/25;
				$blue = floor(255 - $pct*255);
				$green = floor($pct*255);			
			}
			elseif($heatValue < 125) {
				$pct = ($heatValue-25) / 100;
				$green = 255;					
				$red = floor($pct*255);
			}
			else {
				$pct = ($heatValue-125) / 875;
				$green = floor(255 - $pct*255);
				$red = 255;			
			}
			
			$color = imagecolorallocatealpha($img, $red, $green, $blue, 0);
			imagesetpixel($img, $x, $y, $color);
			imagecolordeallocate($img, $color);
		}		
	}

	//Add translucency
	for($y = 0; $y < $height; $y++) {		 
		for($x = 0; $x < $width; $x++) {	
			$old_color = imagecolorat($img, $x, $y);
			$color = imagecolorallocatealpha($img, ($old_color >> 16) & 0xFF, ($old_color >> 8) & 0xFF, $old_color & 0xFF, $alpha);		
			imagesetpixel($img, $x, $y, $color);
			imagecolordeallocate($img, $color);
		}		
	}
	
	//Save the image
	echo 'Saving the image in PNG format to "'.$imgOut.'"...'.PHP_EOL;
	if(!imagepng($img, $imgOut)) die('Could not save the image to "'.$imgOut.'".'.PHP_EOL);
	imagedestroy($img);
	
	echo 'Finished. Memory usage peak during heat map generation: '.(intval(memory_get_peak_usage())/1048576).'MB / '.ini_get('memory_limit').'B'.PHP_EOL;
?>


<?php 
	/**
	* Recursively smoothens the heat map landscape. 
	* Before this function is executed on heat map data, the data is only a collection of individual 
	* pixels which may have high heat values - let's call it altitude. This function turns those 
	* individual pixel "columns" into "hills" by raising the altitude of nearby pixels and calling this 
	* function recursively on them. The values $maxDropPerPixel and $maxDropPerPixelDiag define how big 
	* of an altitude change can be between adjacent pixels, perpendicularly and diagonally. If the drop 
	* is bigger than allowed, the pixel with the lower altitude is raised.
	*/
	function smoothenDrops($x, $y) {
		global $heatMap;
		global $width;
		global $height;
		global $maxDropPerPixel; //North, East, South, West
		global $maxDropPerPixelDiag; //Diagonal

		$x = intval($x);
		$y = intval($y);
		
		if($x > 0) {
			//West		
			if(!isset($heatMap[$x-1][$y])) $heatMap[$x-1][$y] = 0;
			if($heatMap[$x][$y] - $heatMap[$x-1][$y] > $maxDropPerPixel) {
				$heatMap[$x-1][$y] = max(0, $heatMap[$x][$y] - $maxDropPerPixel + 0.01);
				if($heatMap[$x-1][$y] <= 0) unset($heatMap[$x-1][$y]);
				else smoothenDrops($x-1, $y);
			}	
		}

		if($x < $width -1) {
			//East
			if(!isset($heatMap[$x+1][$y])) $heatMap[$x+1][$y] = 0;
			if($heatMap[$x][$y] - $heatMap[$x+1][$y] > $maxDropPerPixel) {
				$heatMap[$x+1][$y] = max(0, $heatMap[$x][$y] - $maxDropPerPixel + 0.01);
				if($heatMap[$x+1][$y] <= 0) unset($heatMap[$x+1][$y]);
				else smoothenDrops($x+1, $y);
			}
		}
		
		if($y > 0) {
			//North	
			if(!isset($heatMap[$x][$y-1])) $heatMap[$x][$y-1] = 0;
			if($heatMap[$x][$y] - $heatMap[$x][$y-1] > $maxDropPerPixel) {
				$heatMap[$x][$y-1] = max(0, $heatMap[$x][$y] - $maxDropPerPixel + 0.01);
				if($heatMap[$x][$y-1] <= 0) unset($heatMap[$x][$y-1]);
				else smoothenDrops($x, $y-1);
			}		
		}
		
		if($y < $height -1) {
			//South
			if(!isset($heatMap[$x][$y+1])) $heatMap[$x][$y+1] = 0;
			if($heatMap[$x][$y] - $heatMap[$x][$y+1] > $maxDropPerPixel) {
				$heatMap[$x][$y+1] = max(0, $heatMap[$x][$y] - $maxDropPerPixel + 0.01);
				if($heatMap[$x][$y+1] <= 0) unset($heatMap[$x][$y+1]);
				else smoothenDrops($x, $y+1);
			}
		}

		
		if($x < $width -1 && $y < $height-1) {
			//South East
			if(!isset($heatMap[$x+1][$y+1])) $heatMap[$x+1][$y+1] = 0;
			if($heatMap[$x][$y] - $heatMap[$x+1][$y+1] > $maxDropPerPixelDiag) {
				$heatMap[$x+1][$y+1] = max(0, $heatMap[$x][$y] - $maxDropPerPixelDiag + 0.01);
				if($heatMap[$x+1][$y+1] <= 0) unset($heatMap[$x+1][$y+1]);
				else smoothenDrops($x+1, $y+1);
			}
		}


		if($x > 0 && $y < $height-1) {
			//South West
			if(!isset($heatMap[$x-1][$y+1])) $heatMap[$x-1][$y+1] = 0;			
			if($heatMap[$x][$y] - $heatMap[$x-1][$y+1] > $maxDropPerPixelDiag) {
				$heatMap[$x-1][$y+1] = max(0, $heatMap[$x][$y] - $maxDropPerPixelDiag + 0.01);
				if($heatMap[$x-1][$y+1] <= 0) unset($heatMap[$x-1][$y+1]);
				else smoothenDrops($x-1, $y+1);
			}		
		}
		
		
		if($x > 0 && $y > 0) {
			//North West
			if(!isset($heatMap[$x-1][$y-1])) $heatMap[$x-1][$y-1] = 0;
			if($heatMap[$x][$y] - $heatMap[$x-1][$y-1] > $maxDropPerPixelDiag) {
				$heatMap[$x-1][$y-1] = max(0, $heatMap[$x][$y] - $maxDropPerPixelDiag + 0.01);
				if($heatMap[$x-1][$y-1] <= 0) unset($heatMap[$x-1][$y-1]);
				else smoothenDrops($x-1, $y-1);
			}
		}


		if($x < $width-1 && $y > 0) {
			//North East
			if(!isset($heatMap[$x+1][$y-1])) $heatMap[$x+1][$y-1] = 0;
			if($heatMap[$x][$y] - $heatMap[$x+1][$y-1] > $maxDropPerPixelDiag) {
				$heatMap[$x+1][$y-1] = max(0, $heatMap[$x][$y] - $maxDropPerPixelDiag + 0.01);
				if($heatMap[$x+1][$y-1] <= 0) unset($heatMap[$x+1][$y-1]);
				else smoothenDrops($x+1, $y-1);
			}
		}
	}
?>


<?
	function getSchemaString() {
		return 
		'<?xml version="1.0" encoding="UTF-8"?>
		<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema" elementFormDefault="qualified">
		  <xs:element name="dataResponse">
			<xs:complexType>
			  <xs:sequence>
				<xs:element ref="locations"/>
			  </xs:sequence>
			</xs:complexType>
		  </xs:element>
		  <xs:element name="locations">
			<xs:complexType>
			  <xs:sequence>
				<xs:element minOccurs= "0" maxOccurs="unbounded" ref="location"/>
			  </xs:sequence>
			</xs:complexType>
		  </xs:element>
		  <xs:element name="location">
			<xs:complexType>
			  <xs:sequence>
				<xs:element ref="latitude"/>
				<xs:element ref="longitude"/>
				<xs:element ref="users"/>
				<xs:element ref="touristPercentage" minOccurs="0" maxOccurs="1"/>
			  </xs:sequence>
			  <xs:attribute name="id" use="optional" type="xs:integer"/>
			</xs:complexType>
		  </xs:element>
		  <xs:element name="latitude" type="xs:decimal"/>
		  <xs:element name="longitude" type="xs:decimal"/>
		  <xs:element name="users" type="xs:integer"/>
		  <xs:element name="touristPercentage" type="xs:decimal"/>
		</xs:schema>';
	}
	
	function getUsageString() {
		return 'Usage: php create_heatmap.php [xml filename] [image width] [image height] [translucency 0..127] [top latitude] [left longitude] [bottom latitude] [right longitude] [image output filename]';
	}
	
?>